'''
check test vectors
'''


from lib1305 import poly1305


def test_vector1() -> None:
    '''
    backported from NaCl library
    '''
    m = b'\x8e\x99\x3b\x9f\x48\x68\x12\x73\xc2\x96\x50\xba\x32\xfc\x76\xce'
    m += b'\x48\x33\x2e\xa7\x16\x4d\x96\xa4\x47\x6f\xb8\xc5\x31\xa1\x18\x6a'
    m += b'\xc0\xdf\xc1\x7c\x98\xdc\xe8\x7b\x4d\xa7\xf0\x11\xec\x48\xc9\x72'
    m += b'\x71\xd2\xc2\x0f\x9b\x92\x8f\xe2\x27\x0d\x6f\xb8\x63\xd5\x17\x38'
    m += b'\xb4\x8e\xee\xe3\x14\xa7\xcc\x8a\xb9\x32\x16\x45\x48\xe5\x26\xae'
    m += b'\x90\x22\x43\x68\x51\x7a\xcf\xea\xbd\x6b\xb3\x73\x2b\xc0\xe9\xda'
    m += b'\x99\x83\x2b\x61\xca\x01\xb6\xde\x56\x24\x4a\x9e\x88\xd5\xf9\xb3'
    m += b'\x79\x73\xf6\x22\xa4\x3d\x14\xa6\x59\x9b\x1f\x65\x4c\xb4\x5a\x74'
    m += b'\xe3\x55\xa5'
    k = b'\xee\xa6\xa7\x25\x1c\x1e\x72\x91\x6d\x11\xc2\xcb\x21\x4d\x3c\x25'
    k += b'\x25\x39\x12\x1d\x8e\x23\x4e\x65\x2d\x65\x1f\xa4\xc8\xcf\xf8\x80'
    a = b'\xf3\xff\xc7\x70\x3f\x94\x00\xe5\x2a\x7d\xfb\x4b\x3d\x33\x05\xd9'
    poly1305.verify(a, m, k)
    assert a == poly1305.auth(m, k)


def test_vector2() -> None:
    '''
    This function primarily tests the internal freeze() function in lib1305.
    The freeze() function detects if the number is less than p (2**130-5) and
    is therefore finally reduced and if not, then p (2**130-5) is subtracted.
    And this test tests the rare condition where the number is greater than 'p'
    and needs to be subtracted.
    '''
    m = 16 * b'\xff'
    k = b'\x02' + 31 * b'\x00'
    a = b'\x03' + 15 * b'\x00'
    poly1305.verify(a, m, k)
    assert a == poly1305.auth(m, k)



'''
Python wrapper around implementation of the Poly1305 one-time authenticator.

Import library:

    from lib1305 import poly1305

Authenticating a message:

    a = poly1305.auth(m, k)

Verifying an authenticator:

    poly1305.verify(a, m, k)

The poly1305.auth function generates an 128-bit authenticator 'a' given
a message 'm' and a 256-bit secret key 'k'. "One-time" means that the secret
key must not be reused to generate an authenticator of another message.

The poly1305.verify function verifies an 128-bit authenticator 'a' given
a messagea 'm' and a 256-bit secret key 'k'.
It raises an exception if the authenticator is not valid.
'''

from .onetimeauth import poly1305



'''
Import shared library.
'''


from ctypes import CDLL as _CDLL
from ctypes.util import find_library as _find_library

_libname = _find_library('1305')
if _libname is None:
    raise FileNotFoundError("unable to locate library '1305'")
_lib = _CDLL(_libname)


def _check_input(x, xlen, name):
    if not isinstance(x, bytes):
        raise TypeError(f'{name} must be bytes')
    if xlen != -1 and xlen != len(x):
        raise ValueError(f'{name} length must have exactly {xlen} bytes')



'''
random data test
'''


import os
from secrets import randbelow
from lib1305 import poly1305


def test_random() -> None:
    '''
    random data test
    '''

    # generate a random 256-bit secret key
    k = os.urandom(32)

    # generate a random massage
    m = os.urandom(128 + randbelow(128))

    # compute an 128-bit authenticator
    a = poly1305.auth(m, k)

    # verify and 128-bit authenticator
    poly1305.verify(a, m, k)

    # replace byte in a message and check if verification fails
    for i in range(len(m)):
        m1 = bytearray(m)
        m1[i] = (m1[i] + 1 + randbelow(255)) % 256
        m1 = bytes(m1)

        try:
            poly1305.verify(a, m1, k)
        except ValueError:
            pass
        else:
            raise ValueError('message forgery not detected !!!')

    # replace byte in an authentificatior and check if verification fails
    for i in range(len(a)):
        a1 = bytearray(a)
        a1[i] = (a1[i] + 1 + randbelow(255)) % 256
        a1 = bytes(a1)

        try:
            poly1305.verify(a1, m, k)
        except ValueError:
            pass
        else:
            raise ValueError('authenticator forgery not detected !!!')



'''
Onetimeauth: secret-key single-message authentication module.
'''

from typing import Tuple as _Tuple
import ctypes as _ct
from ._lib import _lib, _check_input


class Poly1305:
    '''
    Poly1305 one-time authenticator.
    '''

    KEYBYTES = 32
    BYTES = 16

    def __init__(self) -> None:
        '''
        '''
        self._c_auth = getattr(_lib, 'lib1305_onetimeauth_poly1305')
        self._c_auth.argtypes = [_ct.c_char_p,
                                 _ct.c_char_p, _ct.c_longlong, _ct.c_char_p]
        self._c_auth.restype = None
        self._c_verify = getattr(_lib, 'lib1305_onetimeauth_poly1305_verify')
        self._c_verify.argtypes = [_ct.c_char_p,
                                   _ct.c_char_p, _ct.c_longlong, _ct.c_char_p]
        self._c_verify.restype = _ct.c_int

    def auth(self, m: bytes, k: bytes) -> bytes:
        '''
        Auth - generates an authenticator 'a' given a message 'm' and a secret key 'k'.

        Parameters:
            m (bytes): message
            k (bytes): secret key

        Returns:
            a (bytes): authenticator
        '''
        _check_input(m, -1, 'm')
        _check_input(k, self.KEYBYTES, 'k')
        mlen = _ct.c_longlong(len(m))
        m = _ct.create_string_buffer(m)
        k = _ct.create_string_buffer(k)
        a = _ct.create_string_buffer(self.BYTES)
        self._c_auth(a, m, mlen, k)
        return a.raw

    def verify(self, a: bytes, m: bytes, k: bytes) -> None:
        '''
        Verify - verifies an authenticator 'a' given a message 'm' and a secret key 'k'.

        Parameters:
            a (bytes): authenticator
            m (bytes): message
            k (bytes): secret key

        Raises:
            ValueError: an authenticator doesn't match

        Returns:
            None
        '''
        _check_input(a, self.BYTES, 'a')
        _check_input(m, -1, 'm')
        _check_input(k, self.KEYBYTES, 'k')
        mlen = _ct.c_longlong(len(m))
        m = _ct.create_string_buffer(m)
        k = _ct.create_string_buffer(k)
        a = _ct.create_string_buffer(a)
        if self._c_verify(a, m, mlen, k):
            raise ValueError('verify failed')


poly1305 = Poly1305()
